from __future__ import annotations
import torch
import os, re
from tokenizers import Tokenizer as HFTokenizer, models
from ..util import synchronized
from ..util.file import maybe_read_json
from ..model.config import Config
from functools import lru_cache, cached_property
from typing import TYPE_CHECKING
from ..util import profile_opt
if TYPE_CHECKING:
    from . import MMEmbedding

os.environ["TOKENIZERS_PARALLELISM"] = "false"

class Tokenizer:

    def __init__(
        self,
        config: Config,
    ):
        self.config = config

        # TODO: Profile/optimize

        # Defaults
        self.unk_token = "<unk>"
        self.bos_token = "<s>"
        self.eos_token = "</s>"
        self.pad_token = ""
        self.newline_token = "\n"
        self.space_token = " "
        self.special_delimiters = None
        self.unspecial_delimiters = None
        self.extended_piece_to_id = {}
        self.extended_id_to_piece = {}
        self.unspecial_piece_to_id = {}
        self.unspecial_id_to_piece = {}
        self.vocab = None
        self.hf_tokenizer = None

        # Regex
        self.ord_exp = re.compile(r"^<0x([0-9A-Fa-f]+)>$")

        # Files
        self.path_tokenizer_json = os.path.join(self.config.directory, "tokenizer.json")
        self.path_tokenizer_config_json = os.path.join(self.config.directory, "tokenizer_config.json")
        self.path_added_tokens_json = os.path.join(self.config.directory, "added_tokens.json")
        self.tokenizer = HFTokenizer.from_file(self.path_tokenizer_json)
        self.tokenizer_config_dict = maybe_read_json(self.path_tokenizer_config_json)
        self.added_tokens_dict = maybe_read_json(self.path_added_tokens_json)

        # Disable truncation
        self.tokenizer.no_truncation()

        # Deduce placeholders used for space and newline chars in raw vocabulary
        self.space_char_ = " "
        self.newline_char_ = "\n"
        m = self.tokenizer.model
        if isinstance(m, models.BPE):
            def deduce_char_map(input_char):
                char_id = self.tokenizer.encode(input_char, add_special_tokens = False).ids[-1]
                char_str = self.tokenizer.id_to_token(char_id)
                match = self.ord_exp.match(char_str)
                if match:
                    h = match.group(1)
                    o = int(h, 16)
                    char_str = chr(o)
                else:
                    char_str = char_str[-1]
                return char_str
            self.space_char_ = deduce_char_map(" ")  # "Ġ"
            self.newline_char_ = deduce_char_map("\n")  # "Ċ"

        # Add tokens from added_tokens.json if present, assume they're all special
        self.extended_piece_to_id.update(self.added_tokens_dict)

        # Add special tokens from tokenizer_config.json
        atd = self.tokenizer_config_dict.get("added_tokens_decoder", {})
        for (k, v) in atd.items():
            token_id = int(k)
            token_str = v["content"]
            self.extended_piece_to_id[token_str] = token_id
            if not v["special"]:
                self.unspecial_piece_to_id[token_str] = token_id

        # Remove unspecial added tokens that exist in the base tokenizer already, but only if they decode correctly
        # see https://github.com/huggingface/tokenizers/issues/1392
        ok_tokens = []
        for p, i in self.unspecial_piece_to_id.items():
            try:
                itp = self.tokenizer.decode([i], skip_special_tokens = False)
                if itp == p: ok_tokens.append(p)
            except IndexError:
                pass
        for t in ok_tokens: del self.unspecial_piece_to_id[t]

        # Invert extended dictionaries
        self.extended_id_to_piece = {v: k for k, v in self.extended_piece_to_id.items()}
        self.unspecial_id_to_piece = {v: k for k, v in self.unspecial_piece_to_id.items()}

        # Get control token IDs
        ut = self.tokenizer.model.unk_token
        self.unk_token_id = None if ut is None else self.tokenizer.token_to_id(ut)
        self.eos_token_id = self.config.eos_token_id
        self.bos_token_id = self.config.bos_token_id
        self.pad_token_id = self.config.pad_token_id

        # If model config doesn't specify BOS and EOS tokens, try to load from tokenizer config
        def get_default_token_id(config_key: str, current: int | None, default: int | None):
            if current is not None: return current
            if self.tokenizer_config_dict is not None and config_key in self.tokenizer_config_dict:
                st = self.tokenizer_config_dict[config_key]
                if st is None: return None
                if isinstance(st, dict):
                    stc: str | None = st.get("content", None)
                    if stc is None:
                        return None
                    else:
                        return self.tokenizer.token_to_id(stc)
                elif isinstance(st, str):
                    return self.tokenizer.token_to_id(st)
                else:
                    return None
            else:
                return default
        self.pad_token_id = get_default_token_id("pad_token", self.pad_token_id, None)
        self.bos_token_id = get_default_token_id("bos_token", self.bos_token_id, 1)
        self.eos_token_id = get_default_token_id("eos_token", self.eos_token_id, 2)

        # Get control token strings
        self.unk_token = self.tokenizer.model.unk_token
        self.bos_token = None if self.bos_token_id is None else \
            (self.extended_id_to_piece.get(self.bos_token_id) or self.tokenizer.id_to_token(self.bos_token_id))
        self.eos_token = None if self.eos_token_id is None else \
            (self.extended_id_to_piece.get(self.eos_token_id) or self.tokenizer.id_to_token(self.eos_token_id))

        # Use "<pad>" or BOS token as fallback for padding token
        if self.pad_token_id is None:
            pad_test = self.tokenizer.token_to_id("<pad>")
            if pad_test:
                self.pad_token_id = pad_test
            elif self.eos_token_id != self.bos_token_id:
                self.pad_token_id = self.eos_token_id
            else:
                self.pad_token_id = -1

        # Special case if <unk> and <pad> have the same ID
        if self.unk_token_id == self.pad_token_id:
            self.unk_token = self.pad_token

        # Make sure extended vocab contains control tokens, but avoid empty pieces
        if self.unk_token:
            self.extended_piece_to_id[self.unk_token] = self.unk_token_id
            self.extended_id_to_piece[self.unk_token_id] = self.unk_token
        if self.bos_token:
            self.extended_piece_to_id[self.bos_token] = self.bos_token_id
            self.extended_id_to_piece[self.bos_token_id] = self.bos_token
        if self.eos_token:
            self.extended_piece_to_id[self.eos_token] = self.eos_token_id
            self.extended_id_to_piece[self.eos_token_id] = self.eos_token

        self.actual_vocab_size = 1 + max(
            list(self.extended_id_to_piece.keys()) + \
            list(self.unspecial_id_to_piece.keys()) + \
            [self.raw_vocab_size - 1]
        )

        # Useful token IDs
        try:
            self.newline_token_id = self.tokenizer.encode(self.newline_token, add_special_tokens = False).ids[-1]
        except:
            self.newline_token_id = None
        try:
            self.space_token_id = self.tokenizer.encode(self.space_token, add_special_tokens = False).ids[-1]
        except:
            self.space_token_id = None

        # Dictionaries and lists
        self.id_to_ord = None
        self.id_to_piece = None
        self.id_to_piece_with_special = None
        self.piece_to_id = None
        self.get_id_to_ord_list()
        self.get_id_to_piece_list(True)
        self.get_id_to_piece_list(False)
        self.get_piece_to_id_dict()

    @cached_property
    def raw_vocab_size(self):
        """
        Cache this function because it's suspiciously slow in HF Tokenizers
        """
        return self.tokenizer.get_vocab_size()

    def single_token(self, token_id: int) -> torch.Tensor:
        """
        Get single token as tensor

        :param token_id:
            Token ID

        :return:
            LongTensor of shape (1, 1) of token ID
        """
        return torch.tensor([[token_id]], dtype = torch.long)

    def single_id(self, token: str) -> int:
        """
        Get the ID of a single token from exact string match

        :param token:
            Token

        :return:
            int
        """
        tid = self.extended_piece_to_id.get(token, self.get_piece_to_id_dict().get(token))
        return tid

    # Encode string with added, unspecial tokens
    def encode_part(self, text: str, special: bool) -> list[int]:
        if not self.unspecial_piece_to_id:
            return self.tokenizer.encode(text, add_special_tokens = special).ids

        if self.unspecial_delimiters is None:
            self.unspecial_delimiters = re.compile(
                "(" + "|".join(map(re.escape, self.unspecial_piece_to_id.keys())) + ")")

        split = self.unspecial_delimiters.split(text)
        encoded = []

        i = 0
        while i < len(split):
            if split[i] != "": encoded += self.tokenizer.encode(split[i], add_special_tokens = special).ids
            if i + 1 < len(split): encoded += [self.unspecial_piece_to_id[split[i + 1]]]
            i += 2

        return encoded

    # Encode string with special tokens

    def encode_special_or_unspecial(
        self,
        text: str,
        special: bool,
        embeddings: list[MMEmbedding]
    ):
        out_parts = []

        if embeddings:
            aliases = {e.text_alias: e for e in embeddings}
            split_pattern = r"(" + "|".join(re.escape(k) for k in aliases.keys()) + ")"
            in_parts = re.split(split_pattern, text)
        else:
            aliases = {}
            in_parts = [text]

        for text in in_parts:
            if text in aliases:
                out_parts += aliases[text].token_list
            else:
                out_parts += self.encode_part(text, special)

        return out_parts

    # Encode string or list of strings

    def encode(
        self,
        text: str | list[str],
        add_bos: bool = False,
        add_eos: bool = False,
        encode_special_tokens: bool = False,
        return_offsets: bool = False,
        embeddings: list[MMEmbedding] | None = None
    ) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:

        """
        Encode string or list of strings

        :param text:
            str or list[str]: Input text

        :param add_bos:
            Prepend BOS token before each sequence

        :param add_eos:
            Append EOS token to each sequence

        :param encode_special_tokens:
            Encode any special tokens that appear in the input as such. If False, substrings like "<bos>"
            will be encoded as text.

        :param return_offsets:
            Also return a tensor of padding lengths

        :param embeddings:
            List of MMEmbeddings. If present, aliases in the input will be replaced with token ID ranges.

        :return:
            Tensor of shape (batch_size, max_seq_len), optionally offsets Tensor of shape (batch_size)
        """

        if embeddings is None:
            embeddings = []

        if isinstance(text, list):

            # text is a list of strings

            list_ids = [self.encode_special_or_unspecial(t, encode_special_tokens, embeddings) for t in text]

            if add_bos and self.bos_token_id is not None:
                for ids in list_ids: ids.insert(0, self.bos_token_id)
            if add_eos and self.eos_token_id is not None:
                for ids in list_ids: ids.append(self.eos_token_id)

            max_length = max([len(ids) for ids in list_ids])

            padded_ids = []
            offsets = []
            for ids in list_ids:
                padding_length = max_length - len(ids)
                padding = torch.full((padding_length,), self.pad_token_id)
                padded_ids.append(torch.cat((padding, torch.tensor(ids)), dim=0))
                offsets.append(-padding_length)

            stacked_ids = torch.stack(padded_ids, dim=0)

            if return_offsets:
                return stacked_ids, torch.tensor(offsets, dtype=torch.int)
            else:
                return stacked_ids

        else:

            # text is a single string

            # ids = self.encode_special(text) if encode_special_tokens else self.encode_unspecial(text)
            ids = self.encode_special_or_unspecial(text, encode_special_tokens, embeddings)
            if add_bos and self.bos_token_id is not None:
                ids.insert(0, self.bos_token_id)
            if add_eos and self.eos_token_id is not None:
                ids.append(self.eos_token_id)

            ids = torch.tensor(ids).to(torch.long).unsqueeze(0)
            if return_offsets:
                return ids, torch.tensor([0], dtype=torch.int)
            else:
                return ids

    # Decode sequence with added, unspecial tokens

    def decode_unspecial(self, seq):

        if not self.unspecial_id_to_piece:
            return self.tokenizer.decode(seq)

        text = ""
        start = 0
        end = 0
        while end < len(seq):
            if seq[end] in self.unspecial_id_to_piece:
                if end > start: text += self.tokenizer.decode(seq[start: end])
                text += self.unspecial_id_to_piece[seq[end]]
                end += 1
                start = end
            else:
                end += 1
        if end > start: text += self.tokenizer.decode(seq[start: end])
        return text

    # Decode sequence with or without special tokens

    def decode_(self, seq, decode_special_tokens):

        if not decode_special_tokens:

            max_token = self.raw_vocab_size
            seq = [t for t in seq if (t != self.pad_token_id and t < max_token and t != self.eos_token_id)]
            if self.eos_token_id in seq: seq = seq[:seq.index(self.eos_token_id)]
            return self.decode_unspecial(seq)

        else:

            max_token = self.raw_vocab_size
            seq = [t for t in seq if (t != self.pad_token_id and t < max_token)]
            text = ""
            start = 0
            end = 0
            while end < len(seq):
                if seq[end] in self.extended_id_to_piece:
                    if end > start: text += self.tokenizer.decode(seq[start: end])
                    text += self.extended_id_to_piece[seq[end]]
                    end += 1
                    start = end
                else:
                    end += 1
            if end > start: text += self.decode_unspecial(seq[start: end])

        return text

    # Decode IDs, or a list of IDs

    def decode(
        self,
        ids: torch.Tensor | list[torch.Tensor],
        decode_special_tokens: bool = False
    ) -> str | list[str]:
        """
        Decode IDs or list of IDs

        :param ids:
            Tensor of ids, shape (batch_size, max_seq_len)

        :param decode_special_tokens:
            Decode special tokens in the input to their text representation. If False, tokens such as
            "<bos>" will become empty strings in the output.

        :return:
            str if batch_size == 1, else list[str]
        """

        if isinstance(ids, list):

            texts = []
            for i in ids:
                texts.append(self.decode(i, decode_special_tokens))
            return texts

        assert isinstance(ids, torch.Tensor), "ids must be Tensor"

        if ids.dim() > 1:

            texts = []
            for i in range(ids.shape[0]):
                seq = ids[i].tolist()
                texts.append(self.decode_(seq, decode_special_tokens))
            return texts

        else:

            ids = ids.tolist()
            text = self.decode_(ids, decode_special_tokens)
            return text

    # Create padding mask

    def padding_mask(self, ids: torch.Tensor) -> torch.Tensor:
        """
        Create padding mask corresponding to IDs tensor

        :param ids:
            Token IDs

        :return:
            Additive bias Tensor of -inf where ids == pad_token_id, 0 elsewhere
        """

        mask_bool = (ids == self.pad_token_id)
        mask = mask_bool.int()
        mask *= -65505 * 2
        mask = mask.half()
        return mask

    def num_tokens(self, text):
        """
        Measure tokenized length of text

        :param text:
            Input text

        :return:
            Tokenized length of text
        """

        ids = self.tokenizer.encode(text)
        return len(ids)

    # Get ordinals of single-byte tokens

    @cached_property
    def _get_id_to_ord_list(self):

        self.id_to_ord = list(range(self.raw_vocab_size))

        def clean_special_chars(p):
            p = p.replace(self.space_char_, " ")
            p = p.replace(self.newline_char_, "\n")
            return p

        def piece_to_ord(p):
            match = self.ord_exp.match(p)
            if match:
                h = match.group(1)
                return int(h, 16)
            if len(p) == 1:
                p = clean_special_chars(p)
                o = ord(p)
                if o <= 255: return o
            return -1

        i = self.raw_vocab_size
        while True:
            if i in self.extended_id_to_piece:
                self.id_to_ord.append(piece_to_ord(self.extended_id_to_piece[i]))
            elif i in self.unspecial_id_to_piece:
                self.id_to_ord.append(piece_to_ord(self.unspecial_id_to_piece[i]))
            elif i < self.actual_vocab_size:
                self.id_to_ord.append(-1)
            else:
                break
            i += 1

        return self.id_to_ord
    @synchronized
    def get_id_to_ord_list(self):
        return self._get_id_to_ord_list

    # Copy vocabulary from model

    @cached_property
    def _get_fixed_vocab(self):
        test_enc = self.tokenizer.encode(" t", add_special_tokens = False)
        test_count = len(test_enc.ids)
        assert test_count > 0, "Tokenizer error, test string encodes to zero tokens"
        test_id = test_enc.ids[0]
        test_piece = self.tokenizer.decode([test_id])

        if test_count == 1 and len(test_piece) == len(" t"):
            vocab = self.tokenizer.decode_batch(
                [[i] for i in range(self.raw_vocab_size)],
                skip_special_tokens = False
            )
        else:
            prefix_id = self.tokenizer.encode(" ", add_special_tokens = False).ids[0]
            prefix_piece = self.tokenizer.decode([prefix_id])
            prefix_len = len(prefix_piece)
            vocab = self.tokenizer.decode_batch(
                [[prefix_id, i] for i in range(self.raw_vocab_size)]
            )
            vocab = [v[prefix_len:] for v in vocab]

        return vocab
    def get_fixed_vocab(self):
        return self._get_fixed_vocab

    @cached_property
    def _get_id_to_piece_list_spc(self, include_special_tokens = False):

        id_to_piece_extended = self.get_id_to_piece_list().copy()
        for k, v in self.extended_id_to_piece.items():
            id_to_piece_extended[k] = v

        self.id_to_piece_with_special = id_to_piece_extended
        return self.id_to_piece_with_special
    @cached_property
    def _get_id_to_piece_list_nonspc(self, include_special_tokens = False):

        self.id_to_piece = self.get_fixed_vocab()

        i = self.raw_vocab_size
        while True:
            if i in self.extended_id_to_piece:
                self.id_to_piece.append(self.extended_id_to_piece[i])
            elif i in self.unspecial_id_to_piece:
                self.id_to_piece.append(self.unspecial_id_to_piece[i])
            elif i < self.actual_vocab_size:
                self.id_to_piece.append("��_undefined_token_��")
            else:
                break
            i += 1

        return self.id_to_piece
    @synchronized
    def get_id_to_piece_list(self, include_special_tokens = False):
        if include_special_tokens:
            return self._get_id_to_piece_list_spc
        else:
            return self._get_id_to_piece_list_nonspc

    @cached_property
    def _get_piece_to_id_dict(self):
        all_pieces = self.get_id_to_piece_list()
        self.piece_to_id = {piece: idx for idx, piece in enumerate(all_pieces)}
        return self.piece_to_id
    @synchronized
    def get_piece_to_id_dict(self):
        return self._get_piece_to_id_dict

    @staticmethod
    def from_config(config: Config):
        return Tokenizer(config)

    @lru_cache(1000)
    def get_tokens_with_prefix_string(self, prefix: str):
        """
        Return list of token IDs for pieces that start with the given prefix
        """
        id_to_piece = self.get_id_to_piece_list()
        tokens = [idx for idx, piece in enumerate(id_to_piece) if piece.startswith(prefix)]
        return tokens

    @lru_cache(1000)
    def get_tokens_with_prefix_id(self, prefix_id: int):
        """
        Return list of token IDs for pieces that start with the given prefix
        """
        id_to_piece = self.get_id_to_piece_list()
        prefix = id_to_piece[prefix_id]
        return self.get_tokens_with_prefix_string(prefix)

    @cached_property
    def _get_vocab_dict(self):
        """
        Return tokenizer (dictionary for Formatron)
        """
        return {
            self.tokenizer.id_to_token(i): i
            for i in range(self.tokenizer.get_vocab_size())
        }
    def get_vocab_dict(self):
        return self._get_vocab_dict

    def hf_chat_template(
        self,
        messages: list,
        add_generation_prompt: bool = True
    ):
        """
        Tokenize with HF tokenizer. Requires `transformers`

        :param messages:
            HF-formatted list of messages, e.g.
                [{
                    "role": "user",
                    "content": "Hello."
                }]

        :param add_generation_prompt:
            bool, add generation prompt

        :return:
            Topken IDs tensor, shape (1, num_tokens)
        """

        from transformers import AutoTokenizer
        if self.hf_tokenizer is None:
            self.hf_tokenizer = AutoTokenizer.from_pretrained(self.config.directory)
        ids = self.hf_tokenizer.apply_chat_template(
            messages,
            add_generation_prompt = add_generation_prompt
        )["input_ids"]
        ids = torch.tensor(ids, dtype = torch.long).unsqueeze(0)
        return ids



from __future__ import annotations
import torch
import torch.distributed as dist
import time
import numpy as np
from .model_tp_cuda import cuda_host_register, cuda_host_unregister, CUDA_HOST_REGISTER_PORTABLE
from ..ext import exllamav3_ext as ext
from multiprocessing import shared_memory
from ..util import log_tp

GLOBALS_SIZE = 128*1024
SHBUF_SIZE = 16 * 1024 ** 2
SHBUF_SIZE_R = 17 * 128 * 1024
SHBUF_SIZE_S = 16 * 1024
MAX_CPU_REDUCE = SHBUF_SIZE_R // 17 // 256 * 256

class TPBackend:

    def __init__(self):
        pass

    def close(self):
        pass

    def fwd_barrier(self):
        raise NotImplementedError()


class TPBackendNCCL:

    def __init__(
        self,
        device: int,
        active_devices: list[int],
        output_device: int,
        init_method: str,
        master: bool,
        uuid: str,
        shbuf_size: int = SHBUF_SIZE,
    ):
        self.device = device
        if device < 0:
            log_tp(device, f"NCCL init: skip CPU process")
            return

        self.active_devices = active_devices
        self.world_size = len(active_devices)
        self.rank = active_devices.index(device)

        log_tp(device, f"NCCL init: world_size {self.world_size}, rank {self.rank}, device {device}, init_method {init_method}")
        print(f" -- NCCL init: world_size {self.world_size}, rank {self.rank}, device {device}, init_method {init_method}")
        dist.init_process_group(
            "nccl",
            rank = self.rank,
            world_size = self.world_size,
            init_method = init_method,
        )
        self.mp_warmup_nccl(device)
        self.fallback = TPBackendNative(
            device,
            active_devices,
            output_device,
            init_method,
            master,
            uuid,
            shbuf_size
        )


    def mp_warmup_nccl(self, device):
        """
        NCCL does lazy initialization which causes the first reduction operation to take an exceedingly long time
        (20+ seconds). This seems to lead to race conditions or timeouts if it happens during a forward pass. Called
        by TP loader as soon as processes are spawned and process group is initialized.
        """
        print(f" -- NCCL warmup, device {device}, please wait...")
        x = torch.ones((6,), device = device)
        dist.all_reduce(x)
        print(f" -- Finished NCCL warmup, device {device}")


    def close(self):
        if self.device < 0:
            log_tp(self.device, f"NCCL close: skip CPU process")
            return

        dist.barrier()
        self.fallback.close()
        dist.destroy_process_group()


    def fwd_barrier(self):
        dist.barrier()


    def broadcast(self, tensor: torch.Tensor, src_device: int):
        self.fallback.broadcast(tensor, src_device)
        # src_rank = self.active_devices.index(src_device)
        # dist.broadcast(tensor, src = src_rank)


    def all_reduce(self, tensor: torch.Tensor, contribution: bool = True):
        if tensor.dtype == torch.float32:
            temp = tensor.to(torch.bfloat16)
            dist.all_reduce(temp, async_op = False)
            temp = temp.to(torch.float32)
            tensor.copy_(temp)
        else:
            dist.all_reduce(tensor, async_op = False)


    def gather(
        self,
        tensor: torch.Tensor,
        out_tensor: torch.Tensor | None,
        gather_devices: torch.Tensor | None,
        out_device: int,
        ldims: list[int]
    ):
        self.fallback.gather(tensor, out_tensor, gather_devices, out_device, ldims)
        # dst_rank = self.active_devices.index(out_device)
        # d_ldims = [0] * (max(self.active_devices) + 1)
        # for d, m in zip(gather_devices, ldims):
        #     d_ldims[d] = m
        # ldims = [d_ldims[d] for d in self.active_devices]
        #
        # if self.rank == dst_rank:
        #     od = 0
        #     for src, ldim in enumerate(ldims):
        #         if ldim == 0:
        #             continue
        #         out_slice = out_tensor[..., od : od + ldim]
        #         od += ldim
        #         if src == self.rank:
        #             out_slice.copy(tensor)
        #         else:
        #             # print(f"rank {self.rank} recv {out_slice.shape[-1]} from {src}")
        #             rbuf = torch.empty_like(out_slice)
        #             dist.recv(rbuf, src = src)
        #             out_slice.copy_(rbuf)
        # elif tensor.shape[-1] > 0:
        #     # print(f"rank {self.rank} send {tensor.shape[-1]} to {dst_rank}")
        #     dist.send(tensor, dst = dst_rank)


    def run_cpu_reduce_jobs(self):
        pass


    def end_cpu_reduce_jobs(self):
        pass


class TPBackendNative:

    def __init__(
        self,
        device: int,
        active_devices: list[int],
        output_device: int,
        init_method: str,
        master: bool,
        uuid: str,
        shbuf_size: int = SHBUF_SIZE,
        cpu: bool = False
    ):
        self.uuid = uuid
        self.shm_g_name = uuid + "_g"
        self.shm_b_name = uuid + "_b"
        self.shm_r_name = uuid + "_r"
        self.shm_s_name = uuid + "_s"
        self.device = device
        self.max_num_devices = max(active_devices) + 1
        self.active_devices = active_devices
        self.shbuf_size = shbuf_size
        self.master = master
        self.cpu = cpu
        self.cpu_is_pinned = False

        size_g = GLOBALS_SIZE
        size_b = self.shbuf_size
        size_r = SHBUF_SIZE_R
        size_s = SHBUF_SIZE_S

        if master:
            log_tp(device, f"Creating SHMs")
            self.shm_g = shared_memory.SharedMemory(create = True, size = size_g, name = self.shm_g_name)
            log_tp(device, f"Created SHM: {self.shm_g_name}, {size_g} bytes")
            self.shm_b = shared_memory.SharedMemory(create = True, size = size_b, name = self.shm_b_name)
            log_tp(device, f"Created SHM: {self.shm_b_name}, {size_b} bytes")
            self.shm_r = shared_memory.SharedMemory(create = True, size = size_r, name = self.shm_r_name)
            log_tp(device, f"Created SHM: {self.shm_r_name}, {size_r} bytes")
            self.shm_s = shared_memory.SharedMemory(create = True, size = size_s, name = self.shm_s_name)
            log_tp(device, f"Created SHM: {self.shm_s_name}, {size_s} bytes")
            self.buf_g = np.ndarray((size_g,), dtype = np.uint8, buffer = self.shm_g.buf)
            self.buf_b = np.ndarray((size_b,), dtype = np.uint8, buffer = self.shm_b.buf)
            self.buf_r = np.ndarray((size_r,), dtype = np.uint8, buffer = self.shm_r.buf)
            self.buf_s = np.ndarray((size_s,), dtype = np.uint8, buffer = self.shm_s.buf)
            self.buf_g[:] = 0
            self.buf_b[: size_b: 4096] = 0
            self.buf_r[:] = 0
            self.buf_s[:] = 0
        else:
            self.shm_g = None
            self.shm_b = None
            self.shm_r = None
            self.shm_s = None
            deadline = time.time() + 15
            log_tp(device, f"Opening SHMs")
            first_fnf = True
            while True:
                try:
                    if self.shm_g is None:
                        self.shm_g = shared_memory.SharedMemory(name = self.shm_g_name)
                        log_tp(device, f"Opened SHM {self.shm_g_name}")
                    if self.shm_b is None:
                        self.shm_b = shared_memory.SharedMemory(name = self.shm_b_name)
                        log_tp(device, f"Opened SHM {self.shm_b_name}")
                    if self.shm_r is None:
                        self.shm_r = shared_memory.SharedMemory(name = self.shm_r_name)
                        log_tp(device, f"Opened SHM {self.shm_r_name}")
                    if self.shm_s is None:
                        self.shm_s = shared_memory.SharedMemory(name = self.shm_s_name)
                        log_tp(device, f"Opened SHM {self.shm_s_name}")
                    break
                except FileNotFoundError:
                    if first_fnf:
                        log_tp(device, f"Waiting for SHM to appear")
                        first_fnf = False
                    if time.time() > deadline:
                        log_tp(device, f"Timeout opening SHM")
                        raise TimeoutError("Timeout waiting for master process to create SHM")
                    time.sleep(0.05)

        # Create local tensors/flags
        if self.device >= 0:
            self.abort_flag = torch.zeros((1,), device = self.device, dtype = torch.int)
        else:
            self.abort_flag = None

        # Create pinned, shared tensors
        def get_local_tensor(shm_buf, _buffer_size):
            np_view = np.ndarray(
                shape = (_buffer_size,),
                dtype = np.uint8,
                buffer = shm_buf,
                offset = 0,
            )
            return torch.as_tensor(np_view)
        self.tensor_g = get_local_tensor(self.shm_g.buf, size_g)
        self.tensor_b = get_local_tensor(self.shm_b.buf, size_b)
        self.tensor_r = get_local_tensor(self.shm_r.buf, size_r)
        self.tensor_s = get_local_tensor(self.shm_s.buf, size_s)
        self.ptr_g = self.tensor_g.data_ptr()
        self.ptr_b = self.tensor_b.data_ptr()
        self.ptr_r = self.tensor_r.data_ptr()
        self.ptr_s = self.tensor_s.data_ptr()
        if not self.cpu:
            log_tp(device, f"Host register G")
            cuda_host_register(self.ptr_g, self.tensor_g.numel(), flags = CUDA_HOST_REGISTER_PORTABLE)
            log_tp(device, f"Host register B")
            cuda_host_register(self.ptr_b, self.tensor_b.numel(), flags = CUDA_HOST_REGISTER_PORTABLE)
            log_tp(device, f"Host register R")
            cuda_host_register(self.ptr_r, self.tensor_r.numel(), flags = CUDA_HOST_REGISTER_PORTABLE)
            log_tp(device, f"Host register S")
            cuda_host_register(self.ptr_s, self.tensor_s.numel(), flags = CUDA_HOST_REGISTER_PORTABLE)

        # Init global context
        if master:
            log_tp(device, f"Initializing global context")
            ext.pg_init_context(self.ptr_g)


    def close(self):
        if not self.cpu:
            log_tp(self.device, f"Host unregister G")
            cuda_host_unregister(self.ptr_g)
            log_tp(self.device, f"Host unregister B")
            cuda_host_unregister(self.ptr_b)
            log_tp(self.device, f"Host unregister R")
            cuda_host_unregister(self.ptr_r)
            log_tp(self.device, f"Host unregister S")
            cuda_host_unregister(self.ptr_s)
        self.shm_g.close()
        log_tp(self.device, f"Closed {self.shm_g_name}")
        self.shm_b.close()
        log_tp(self.device, f"Closed {self.shm_b_name}")
        self.shm_r.close()
        log_tp(self.device, f"Closed {self.shm_r_name}")
        self.shm_s.close()
        log_tp(self.device, f"Closed {self.shm_s_name}")
        if self.master:
            log_tp(self.device, f"Master unlink G")
            self.shm_g.unlink()
            log_tp(self.device, f"Master unlink B")
            self.shm_b.unlink()
            log_tp(self.device, f"Master unlink R")
            self.shm_r.unlink()
            log_tp(self.device, f"Master unlink S")
            self.shm_s.unlink()


    def fwd_barrier(self):
        ext.pg_barrier(self.ptr_g, self.active_devices, self.device, self.abort_flag)


    def broadcast(self, tensor: torch.Tensor, src_device: int):
        if tensor.numel() * tensor.element_size() <= 2048:
            ext.pg_broadcast_ll(
                self.ptr_g,
                self.active_devices,
                self.device,
                src_device,
                tensor,
                self.ptr_s,
                SHBUF_SIZE_S,
                self.abort_flag
            )
        else:
            ext.pg_broadcast(
                self.ptr_g,
                self.active_devices,
                self.device,
                src_device,
                tensor,
                self.ptr_b,
                self.shbuf_size,
                self.abort_flag
            )


    def all_reduce(self, tensor: torch.Tensor, contribution: bool = True):
        # if tensor.numel() * 2 < MAX_CPU_REDUCE:
        ext.pg_all_reduce_cpu(
            self.ptr_g,
            self.active_devices,
            self.device,
            self.active_devices[0],
            tensor,
            contribution,
            self.ptr_r,
            SHBUF_SIZE_R,
            self.master,
            self.abort_flag
        )
        # else:
        #     ext.pg_all_reduce(
        #         self.ptr_g,
        #         self.active_devices,
        #         self.device,
        #         self.active_devices[0],
        #         tensor,
        #         self.ptr_b,
        #         self.shbuf_size,
        #         self.abort_flag
        #     )


    def gather(
        self,
        tensor: torch.Tensor,
        out_tensor: torch.Tensor | None,
        gather_devices: torch.Tensor | None,
        out_device: int,
        ldims: list[int]
    ):
        if out_device == self.device:
            assert out_tensor is not None, \
                f"Gather: Output device must supply output tensor"
            assert out_tensor.shape[-1] == sum(ldims), \
                f"Gather: Output tensor must match size of concatenated slices: {sum(ldims)}"

        ext.pg_gather(
            self.ptr_g,
            gather_devices,
            self.device,
            out_device,
            tensor,
            out_tensor,
            ldims,
            self.ptr_b,
            self.shbuf_size,
            self.abort_flag
        )


    def run_cpu_reduce_jobs(self):
        # if not self.cpu_is_pinned:
        #     set_process_priority_and_affinity()
        #     self.cpu_is_pinned = True
        ext.run_cpu_reduce_jobs(
            self.ptr_g,
            self.ptr_r,
            SHBUF_SIZE_R,
        )


    def end_cpu_reduce_jobs(self):
        if self.master:
            ext.end_cpu_reduce_jobs(
                self.ptr_g,
            )


from __future__ import annotations
from typing import override
import torch
import torch.nn.functional as F
from ..model.config import Config
from . import Module
from ..util.tensor import get_for_device, to2

class Glm4VPosEmbedding(Module):

    def __init__(
        self,
        config: Config,
        key: str,
        num_position_embeddings: int,
        hidden_size: int,
        spatial_merge_size: int,
        out_dtype: torch.dtype | None = None,
        qmap: str | None = None,
    ):
        super().__init__(config, key, None)
        self.module_name = "Glm4VPosEmbedding"
        self.qmap = qmap
        self.key = key

        self.num_position_embeddings = num_position_embeddings
        self.num_grid_per_side = num_position_embeddings ** 0.5
        self.spatial_merge_size = spatial_merge_size
        self.hidden_size = hidden_size
        self.out_dtype = out_dtype

        self.pos_embed_2d = None


    @override
    def weights_numel(self):
        return self.num_position_embeddings + self.hidden_size


    def optimizer_targets(self):
        raise NotImplementedError()


    @override
    def load(self, device: torch.device, **kwargs):
        self.device = device
        weight = self.config.stc.get_tensor(self.key + ".weight", self.device, allow_bf16 = True, no_defer = True)
        orig_size_sq = weight.shape[0]
        orig_size = int(orig_size_sq**0.5)
        self.pos_embed_2d = (
            weight.view(orig_size, orig_size, self.hidden_size)
            .permute(2, 0, 1)
            .unsqueeze(0)
            .to(device = device, dtype = torch.float32)
        )

    @override
    def unload(self):
        self.device = None
        self.pos_embed_2d = None


    def forward(
        self,
        x: torch.Tensor,
        params: dict,
        out_dtype: torch.dtype | None = None
    ):
        grid_thw = get_for_device(params, "grid_thw", self.device)

        # Position IDs
        spm = self.spatial_merge_size
        t, h, w = grid_thw[0].tolist()
        hpos_ids = torch.arange(h, dtype = torch.float).unsqueeze(1).expand(-1, w)
        hpos_ids = hpos_ids.reshape(h // spm, spm, w // spm, spm)
        hpos_ids = hpos_ids.permute(0, 2, 1, 3)
        hpos_ids = hpos_ids.flatten()
        wpos_ids = torch.arange(w, dtype = torch.float).unsqueeze(0).expand(h, -1)
        wpos_ids = wpos_ids.reshape(h // spm, spm, w // spm, spm)
        wpos_ids = wpos_ids.permute(0, 2, 1, 3)
        wpos_ids = wpos_ids.flatten()
        h_coords = hpos_ids.to(self.device).float()
        w_coords = wpos_ids.to(self.device).float()
        length = x.shape[1]

        # Normalize coordinates to [-1, 1] range for grid_sample
        norm_w = (w_coords + 0.5) / (w / 2)
        norm_h = (h_coords + 0.5) / (h / 2)
        grid = torch.stack((norm_w, norm_h), dim = -1).unsqueeze(0).unsqueeze(2) - 1

        # Perform bicubic interpolation
        interpolated_embed_fp32 = F.grid_sample(
            self.pos_embed_2d,
            grid,
            mode = "bicubic",
            align_corners = False,
            padding_mode = "border"
        )

        # Reshape and convert back to original dtype
        adapted_pos_embed_fp32 = interpolated_embed_fp32.squeeze(0).squeeze(-1).T

        x += adapted_pos_embed_fp32
        return to2(x, out_dtype, self.out_dtype)



"""OptimizedLLM: high-level async inference wrapper.

Provides async generate/stream with vision support.
Internally uses SlimGenerator.
"""

from __future__ import annotations

import asyncio
import functools
import io
import logging
import threading
from collections.abc import AsyncIterator
from pathlib import Path
from typing import TYPE_CHECKING

from huggingface_hub import snapshot_download

if TYPE_CHECKING:
    import torch

    from exllamav3_opt.prefix_cache import PrefixCache

logger = logging.getLogger(__name__)


class LLMConfig:
    """Configuration for OptimizedLLM."""

    def __init__(
        self,
        model_repo: str = "",
        model_revision: str = "bpw3.0",
        max_new_tokens: int = 256,
        cache_size: int = 2048,
        temperature: float = 0.8,
        top_k: int = 50,
        top_p: float = 0.95,
        min_p: float = 0.0,
    ) -> None:
        self.model_repo = model_repo
        self.model_revision = model_revision
        self.max_new_tokens = max_new_tokens
        self.cache_size = cache_size
        self.temperature = temperature
        self.top_k = top_k
        self.top_p = top_p
        self.min_p = min_p


class OptimizedLLM:
    """High-level async inference wrapper using SlimGenerator.

    Prompt formatting uses the model's built-in chat template
    via ``tokenizer.hf_chat_template()``. Sampler and stop
    conditions are derived from config / model metadata.
    """

    def __init__(self, config: LLMConfig, hf_token: str = "") -> None:
        self._config = config
        self._hf_token = hf_token
        self._lock = asyncio.Lock()
        self._loaded = False

        # Set by load()
        self._model = None
        self._tokenizer = None
        self._generator = None  # SlimGenerator
        self._cache = None
        self._vision_model = None
        self._prefix_cache: PrefixCache | None = None
        self._stop_conditions: list[int | str] = []

    def download(self) -> str:
        """Download model from HuggingFace Hub. Returns local path."""
        logger.info(
            "Downloading model %s (rev: %s)",
            self._config.model_repo,
            self._config.model_revision,
        )
        path = snapshot_download(
            self._config.model_repo,
            revision=self._config.model_revision,
            token=self._hf_token or None,
        )
        logger.info("Model downloaded to %s", path)
        return path

    def load(self, model_path: str, enable_prefix_cache: bool = True) -> None:
        """Load ExLlamaV3 model, tokenizer, cache, and SlimGenerator."""
        from exllamav3 import Cache, Config, Model, Tokenizer
        from exllamav3_opt.generator import SlimGenerator
        from exllamav3_opt.prefix_cache import PrefixCache

        logger.info("Loading model from %s", model_path)

        config = Config.from_directory(model_path)
        self._model = Model.from_config(config)
        self._cache = Cache(self._model, max_num_tokens=self._config.cache_size)
        self._model.load(progressbar=True)

        self._tokenizer = Tokenizer.from_config(config)

        # Stop conditions: merge eos_token_id from config.json
        # and generation_config.json (which often has extra stop tokens
        # like <|im_end|> that config.json misses)
        self._stop_conditions = list(config.eos_token_id_list)
        gen_cfg_path = Path(model_path) / "generation_config.json"
        if gen_cfg_path.exists():
            import json

            with open(gen_cfg_path) as f:
                gen_cfg = json.load(f)
            gc_eos = gen_cfg.get("eos_token_id", [])
            if isinstance(gc_eos, int):
                gc_eos = [gc_eos]
            for tid in gc_eos:
                if tid not in self._stop_conditions:
                    self._stop_conditions.append(tid)

        # Prefix cache (optional)
        if enable_prefix_cache:
            self._prefix_cache = PrefixCache()

        # SlimGenerator replaces upstream Generator
        self._generator = SlimGenerator(
            self._model,
            self._cache,
            self._tokenizer,
            prefix_cache=self._prefix_cache,
        )

        # Vision model (separate component, fp16)
        try:
            self._vision_model = Model.from_config(config, component="vision")
            self._vision_model.load(progressbar=True)
            logger.info("Vision model loaded")
        except Exception:
            logger.debug("No vision component found, vision features disabled", exc_info=True)
            self._vision_model = None

        self._loaded = True
        logger.info("Model loaded successfully (OptimizedLLM)")

    @property
    def tokenizer(self):
        return self._tokenizer

    @property
    def vision_model(self):
        return self._vision_model

    def _make_sampler(self):
        """Create sampler from config."""
        from exllamav3.generator.sampler import ComboSampler

        return ComboSampler(
            temperature=self._config.temperature,
            top_k=self._config.top_k,
            top_p=self._config.top_p,
            min_p=self._config.min_p,
        )

    # ------------------------------------------------------------------
    # Prompt building
    # ------------------------------------------------------------------

    def build_prompt(
        self,
        messages: list[dict[str, str]],
    ) -> torch.Tensor:
        """Build token IDs using the model's chat template.

        Args:
            messages: List of {"role": ..., "content": ...} dicts.

        Returns:
            Token IDs tensor (1, seq_len) ready for generate/stream.
        """
        if self._tokenizer is None:
            raise RuntimeError("Model not loaded")
        return self._tokenizer.hf_chat_template(messages, add_generation_prompt=True)

    # ------------------------------------------------------------------
    # Sync internals
    # ------------------------------------------------------------------

    def _generate_sync(
        self,
        input_ids: torch.Tensor,
        max_new_tokens: int | None = None,
        embeddings: list | None = None,
    ) -> str:
        """Synchronous full generation (non-streaming)."""
        if not self._loaded:
            raise RuntimeError("Model not loaded")

        return self._generator.generate(
            input_ids=input_ids,
            stop_conditions=self._stop_conditions,
            sampler=self._make_sampler(),
            max_new_tokens=max_new_tokens or self._config.max_new_tokens,
            embeddings=embeddings,
        )

    def _stream_sync(
        self,
        input_ids: torch.Tensor,
        max_new_tokens: int | None = None,
        embeddings: list | None = None,
        cancel_flag: threading.Event | None = None,
    ):
        """Synchronous streaming generation. Yields token strings."""
        if not self._loaded:
            raise RuntimeError("Model not loaded")

        for chunk in self._generator.stream_tokens(
            input_ids=input_ids,
            max_new_tokens=max_new_tokens or self._config.max_new_tokens,
            sampler=self._make_sampler(),
            stop_conditions=self._stop_conditions,
            embeddings=embeddings,
            cancel_flag=cancel_flag,
        ):
            if chunk.text:
                yield chunk.text

    # ------------------------------------------------------------------
    # Async public API
    # ------------------------------------------------------------------

    async def generate(
        self,
        input_ids: torch.Tensor,
        max_new_tokens: int | None = None,
        embeddings: list | None = None,
    ) -> str:
        """Async full generation with lock."""
        loop = asyncio.get_running_loop()
        async with self._lock:
            return await loop.run_in_executor(
                None,
                functools.partial(
                    self._generate_sync,
                    input_ids,
                    max_new_tokens,
                    embeddings,
                ),
            )

    async def stream(
        self,
        input_ids: torch.Tensor,
        max_new_tokens: int | None = None,
        embeddings: list | None = None,
    ) -> AsyncIterator[str]:
        """Async streaming generation with lock.

        Yields token strings as they are generated.
        Lock is held for the entire stream duration.
        On CancelledError: signals the thread to stop via cancel_flag,
        waits for it to finish, then releases the lock.
        """
        cancel_flag = threading.Event()

        async with self._lock:
            loop = asyncio.get_running_loop()
            token_queue: asyncio.Queue[str | None] = asyncio.Queue()

            def _run():
                try:
                    for token in self._stream_sync(
                        input_ids, max_new_tokens, embeddings, cancel_flag
                    ):
                        loop.call_soon_threadsafe(token_queue.put_nowait, token)
                finally:
                    loop.call_soon_threadsafe(token_queue.put_nowait, None)

            future = loop.run_in_executor(None, _run)

            try:
                while True:
                    token = await token_queue.get()
                    if token is None:
                        break
                    yield token
            except GeneratorExit:
                cancel_flag.set()
            finally:
                await future

    # ------------------------------------------------------------------
    # Vision processing
    # ------------------------------------------------------------------

    def process_image(self, image_bytes: bytes) -> tuple[list, list[str]] | None:
        """Process image for vision input. Returns (embeddings, aliases) or None."""
        if self._vision_model is None:
            return None
        from PIL import Image

        image = Image.open(io.BytesIO(image_bytes)).convert("RGB")
        image.thumbnail((512, 512))
        emb = self._vision_model.get_image_embeddings(self._tokenizer, image)
        return [emb], [emb.text_alias]

    def process_pil_images(self, images) -> tuple[list, list[str]] | None:
        """Process PIL Images for vision input. Returns (embeddings, aliases) or None."""
        if self._vision_model is None or not images:
            return None
        img = images[0].convert("RGB")
        img.thumbnail((512, 512))
        emb = self._vision_model.get_image_embeddings(self._tokenizer, img)
        return [emb], [emb.text_alias]



"""CUDA extension bindings.

Imports the compiled extension if available, otherwise stubs raise ImportError
at call time so the rest of the library works without custom kernels.
"""

try:
    from exllamav3_opt._ext.exllamav3_opt_ext import (  # type: ignore[import-not-found]
        dequant_fp8_cache_paged,
        fused_rmsnorm_residual,
        fused_sample,
        quant_fp8_cache_paged,
    )
except ImportError:
    _MSG = (
        "exllamav3_opt CUDA extension not compiled. "
        "Run: pip install --no-build-isolation -e . "
        "from the exllamav3-optimized directory."
    )

    def fused_rmsnorm_residual(*args, **kwargs):
        raise ImportError(_MSG)

    def fused_sample(*args, **kwargs):
        raise ImportError(_MSG)

    def quant_fp8_cache_paged(*args, **kwargs):
        raise ImportError(_MSG)

    def dequant_fp8_cache_paged(*args, **kwargs):
        raise ImportError(_MSG)


__all__ = [
    "dequant_fp8_cache_paged",
    "fused_rmsnorm_residual",
    "fused_sample",
    "quant_fp8_cache_paged",
]



"""Tests for fused CUDA kernels.

Validates output matches unfused reference implementations.
"""

from __future__ import annotations

import pytest
import torch


@pytest.mark.requires_cuda
class TestFusedRMSNormResidual:

    def test_matches_unfused(self, device):
        """Fused RMSNorm+Residual matches sequential unfused ops."""
        try:
            from exllamav3_opt._ext import fused_rmsnorm_residual
        except ImportError:
            pytest.skip("CUDA extension not compiled")

        dim = 3584  # Qwen3.5-27B hidden size
        rows = 4
        epsilon = 1e-6

        x = torch.randn(rows, dim, dtype=torch.half, device=device)
        attn_out = torch.randn(rows, dim, dtype=torch.half, device=device)
        weight = torch.randn(dim, dtype=torch.half, device=device)

        # Reference: unfused
        x_ref = x.clone()
        x_ref += attn_out
        rms = torch.sqrt(torch.mean(x_ref.float() ** 2, dim=-1, keepdim=True) + epsilon)
        y_ref = (x_ref.float() / rms * weight.float()).half()

        # Fused kernel
        x_fused = x.clone()
        y_fused = torch.empty_like(x_fused)
        fused_rmsnorm_residual(x_fused, attn_out, weight, y_fused, epsilon, 0.0)

        # x should be modified in-place (residual add)
        x_expected = x + attn_out
        assert torch.allclose(x_fused.float(), x_expected.float(), atol=1e-3), \
            f"Residual add mismatch: max diff = {(x_fused - x_expected).abs().max()}"

        # y should match normalized output
        assert torch.allclose(y_fused.float(), y_ref.float(), atol=1e-2), \
            f"RMSNorm output mismatch: max diff = {(y_fused - y_ref).abs().max()}"


@pytest.mark.requires_cuda
class TestFusedSampling:

    def test_argmax_mode(self, device):
        """Temperature=0 (argmax) returns the correct token."""
        try:
            from exllamav3_opt._ext import fused_sample
        except ImportError:
            pytest.skip("CUDA extension not compiled")

        vocab_size = 152064
        logits = torch.randn(1, vocab_size, dtype=torch.half, device=device)
        output_id = torch.zeros(1, dtype=torch.long, device=device)

        fused_sample(logits, output_id, 0.0, 0, 0)

        expected = logits.float().argmax().item()
        assert output_id.item() == expected

    def test_temperature_sampling(self, device):
        """Non-zero temperature produces valid token IDs."""
        try:
            from exllamav3_opt._ext import fused_sample
        except ImportError:
            pytest.skip("CUDA extension not compiled")

        vocab_size = 152064
        logits = torch.randn(1, vocab_size, dtype=torch.half, device=device)
        output_id = torch.zeros(1, dtype=torch.long, device=device)

        fused_sample(logits, output_id, 0.6, 20, 42)

        token_id = output_id.item()
        assert 0 <= token_id < vocab_size

    def test_top_k_constraint(self, device):
        """With top_k=1, should always pick the argmax."""
        try:
            from exllamav3_opt._ext import fused_sample
        except ImportError:
            pytest.skip("CUDA extension not compiled")

        vocab_size = 1000
        logits = torch.randn(1, vocab_size, dtype=torch.half, device=device)
        output_id = torch.zeros(1, dtype=torch.long, device=device)

        # Very high temperature but top_k=1 should still pick argmax
        fused_sample(logits, output_id, 10.0, 1, 123)

        expected = logits.float().argmax().item()
        assert output_id.item() == expected

    def test_deterministic_with_same_seed(self, device):
        """Same seed produces same result."""
        try:
            from exllamav3_opt._ext import fused_sample
        except ImportError:
            pytest.skip("CUDA extension not compiled")

        vocab_size = 152064
        logits = torch.randn(1, vocab_size, dtype=torch.half, device=device)

        output1 = torch.zeros(1, dtype=torch.long, device=device)
        output2 = torch.zeros(1, dtype=torch.long, device=device)

        fused_sample(logits, output1, 0.6, 20, 42)
        fused_sample(logits, output2, 0.6, 20, 42)

        assert output1.item() == output2.item()
